"""
This module contains tests for:

- test_helmreleases: if all the helmreleases in EXPECTED_RELEASES are
  deployed successfully
- test_apps_running: if all the applications are running without problems
  (i.e., condition = Ready for all the related pods)
"""
import os
from kubernetes import client, config
from kubernetes.client.rest import ApiException
import pytest

EXPECTED_KUSTOMIZATIONS_BASE = [
    'openappstack',
    'monitoring',
    'infrastructure',
    'core',
]

EXPECTED_RELEASES = {
    'cert-manager': ['cert-manager'],
    'kube-system': ['local-path-provisioner'],
    'oas': [
        'ingress',
        'kube-prometheus-stack',
        'loki',
        'promtail',
        'eventrouter',
        'single-sign-on'
    ],
    'oas-apps': ['nextcloud', 'rocketchat', 'wordpress'],
}

EXPECTED_APP_LABELS = {
    'cert-manager': {
        'namespace': 'cert-manager',
        'label_selector': 'app.kubernetes.io/instance=cert-manager'
    },
    'eventrouter': {
        'namespace': 'oas',
        'label_selector': 'app=eventrouter'},
    'local-path-provisioner': {
        'namespace': 'kube-system',
        'label_selector': 'app.kubernetes.io/instance=local-path-provisioner'},
    'loki': {
        'namespace': 'oas',
        'label_selector': 'app=loki'},
    'promtail': {
        'namespace': 'oas',
        'label_selector': 'app=promtail'},
    'nextcloud': {
        'namespace': 'oas-apps',
        'label_selector': 'app.kubernetes.io/instance=nc'},
    'kube-prometheus-stack': {
        'namespace': 'oas',
        'label_selector': 'app in (grafana,prometheus)'},
    'rocketchat': {
        'namespace': 'oas-apps',
        'label_selector': 'app.kubernetes.io/instance=rocketchat'},
    'single-sign-on': {
        'namespace': 'oas',
        'label_selector': 'app.kubernetes.io/name in (hydra,hydra-maester,single-sign-on-userbackend,single-sign-on-consent,single-sign-on-userpanel,single-sign-on-login)'},
    'wordpress': {
        'namespace': 'oas-apps',
        'label_selector': 'release=wordpress'}
}


def get_kustomization_status(name, api, namespace="flux-system"):
    """Returns status contition for kustomization `name` in `namespace`"""
    print('Testing %s in namespace %s ...' % (name, namespace), end='')
    try:
        kustomization = api.get_namespaced_custom_object(
            group="kustomize.toolkit.fluxcd.io",
            version="v1beta1",
            plural="kustomizations",
            name=name,
            namespace=namespace
        )
        ks_status = kustomization['status']['conditions'][0]['status']
    except ApiException as ex:
        if ex.status == 404:
            ks_status = 'Not found'
        else:
            raise
        print(f"**** NOT DEPLOYED, status: {ks_status} *****")
    except KeyError:
        ks_status = 'Key error'
        print("Kustomization key error: ")
        print(kustomization)
    return ks_status

def get_release_status(name, namespace, api):
    """Returns release status for release `name` in `namespace`"""
    print('Testing %s in namespace %s ...' % (name, namespace), end='')
    try:
        release = api.get_namespaced_custom_object(
            group="helm.toolkit.fluxcd.io",
            version="v2beta1",
            plural="helmreleases",
            name=name,
            namespace=namespace
        )
        print(release['status'])
        release_status = release['status']['conditions'][0]['status']
        print(release_status)
    except ApiException as ex:
        if ex.status == 404:
            release_status = 'not found'
        else:
            raise
        print("**** NOT DEPLOYED, status: %s *****" % release_status)
    except KeyError:
        release_status = 'Key error'
        print("HelmRelease key error: ")
        print(release)
    return release_status


def check_all_pods_running(pods):
    """
    Loop through all the pods in an API result.
    If a pod does not have an element `status.phase` with the value "Running",
    return False. Otherwise, returns True.

    :param kubernetes.V1Pod[] pods: list of V1Pod elements to check
    """
    ret = True
    for pod in pods:
        print("- {}: {}".format(pod.metadata.name, pod.status.phase))
        if pod.status.phase != "Running":
            ret = False
    return ret


@pytest.fixture(autouse=True)
def run_around_tests():
    """
    Prepare kube config before running a test
    """
    cluster_dir = os.environ.get("CLUSTER_DIR")
    kubeconfig = os.path.join(cluster_dir, 'secrets',
                              'kube_config_cluster.yml')
    config.load_kube_config(config_file=kubeconfig)
    yield

@pytest.mark.app
@pytest.mark.kustomizations
def test_kustomizations(app):
    """
    Checks if all desired Kustomizations installed by weave flux are in
    'deployed' state.
    """
    if app == 'base':
        kustomizations = EXPECTED_KUSTOMIZATIONS_BASE
    else:
        kustomizations = [app]

    custom_objects = client.CustomObjectsApi()

    failed = 0
    print('\n')
    for kustomization in kustomizations:
        ks_status = get_kustomization_status(kustomization, custom_objects)

        if ks_status != 'True':
            failed += 1

    assert failed == 0, f"Error: {failed} kustomizations not 'ready'!"

@pytest.mark.app
@pytest.mark.helmreleases
def test_helmreleases(app):
    """
    Checks if all desired HelmReleases installed by weave flux are in
    'deployed' state.
    """
    if app != 'all':
        apps = [item for sublist in EXPECTED_RELEASES.values()
                for item in sublist]
        assert app in apps, "Error: Unknown app: {}".format(app)

    custom_objects = client.CustomObjectsApi()

    failed = 0
    print('\n')
    for namespace in EXPECTED_RELEASES:
        for app_name in EXPECTED_RELEASES[namespace]:
            if app in (app_name, 'all'):
                app_status = get_release_status(
                    app_name, namespace, custom_objects)
                if app_status != 'True':
                    failed += 1
    assert failed == 0, "Error: {} apps not 'deployed'!".format(failed)


@pytest.mark.app
@pytest.mark.apps_running
def test_apps_running(app):
    """
    Checks if all the pods related to releases in EXPECTED_RELEASES are
    ready
    """

    if app != 'all':
        assert app in EXPECTED_APP_LABELS, "Error: Unknown app: {}".format(app)

    api = client.CoreV1Api()
    failed = 0
    print('\n')
    for app_name, info in EXPECTED_APP_LABELS.items():
        if app in (app_name, 'all'):
            print("{}: ".format(app_name))
            result = api.list_namespaced_pod(
                namespace=info['namespace'],
                label_selector=info['label_selector'],
            )
            if not check_all_pods_running(result.items):
                failed += 1
            print()
    assert failed == 0, "Error: {} apps not 'Running'!".format(failed)
